/*
 * file:       Base.c
 * purpose:    basic functions like common initialization, common interrupts, ...
 * author:     Santiago
 * copyright:  2008 (c) Creative commons - by-nc-sa - some rights reserved
 *             see http://creativecommons.org/licenses/by-nc-sa/3.0/ for details
 */
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "Config.h"
#include "Base.h"
#include "LCD.h"
#include "Stepper.h"


void init(void) {
   /*
    * inputs
    */
/*   
// null is the default, so don't waste code space
   DDR_T1 &= (uint8_t) ~(1 << PIN_T1);
   DDR_T2 &= (uint8_t) ~(1 << PIN_T2);
   DDR_T3 &= (uint8_t) ~(1 << PIN_T3);

   DDR_PUSH_ENCODE  &= (uint8_t) ~(1 << PIN_PUSH_ENCODE);
   DDR_RIGHT_ENCODE &= (uint8_t) ~(1 << PIN_RIGHT_ENCODE);
   DDR_LEFT_ENCODE  &= (uint8_t) ~(1 << PIN_LEFT_ENCODE);
 */
   DDR_STEP_L1 |= 1 << PIN_STEP_L1;
   DDR_STEP_L2 |= 1 << PIN_STEP_L2;
   DDR_STEP_L3 |= 1 << PIN_STEP_L3;
   DDR_STEP_L4 |= 1 << PIN_STEP_L4;

   DDR_LED1 |= 1 << PIN_LED1;
   DDR_LED2 |= 1 << PIN_LED2;
   DDR_BUZZ |= 1 << PIN_BUZZ;

   PORT_LED1 |= 1 << PIN_LED1;
   PORT_LED2 |= 1 << PIN_LED2;

   lcdInit();

   _delay_ms(250);

   /*
    * timer 0 - stepper speed
    */
   stepInit();

   /*
    * timer 1 - time base and pwm
    */
   TCCR1A = 1 << WGM11 | 1 << WGM10;                       // fast PWM mode
   TCCR1B = 1 << WGM13 | 1 << WGM12                        // with OCR1A as top
          | 1 << CS10                                      // prescale: 1 == sysclock 
#ifdef TEST_MODE          
          | 1 << CS12                                      // prescale: 1024
#endif   
            ;
   TIMSK1 = 1 << OCIE1B | 1 << TOIE1;                      // as OCR1A is used as top, 
                                                           // only OCR1B can be used for PWM-compare match
   OCR1A  = TIMER_TOP;
   OCR1B  = 2; //STEP_CURRENT_MAX / 2;

   /*
    * timer 2 - sound
    */
   TCCR2A = 1 << WGM21 | 1 << WGM20;                       // using fast PWM mode with CTC
   TCCR2B = 1 << WGM22                                     // ensures synchronized update of top level 
          | 1 << CS22 | 1 << CS21;                         // prescale of 256 (245 Hz - 20 kHz)
   TIMSK2 = 1 << TOIE2;                                    // enable interrupts
   OCR2A  = 87;                                            // ca. 440 Hz

   sei();                                                  // allow interrupts

   PORT_LED1 &= (uint8_t) ~(1 << PIN_LED1);                // lite end of init
   PORT_LED2 &= (uint8_t) ~(1 << PIN_LED2);
}


/*
 * overflow of timer 1 is time and pwm base
 */
ISR(TIMER1_OVF_vect) {
   static int8_t enc_last = 0x01;
   int8_t i=0;

   if (++work.cIPS >= TIMER_IPS) {
      work.cIPS = 0;
      work.time++;
   }

   /*
    * pwm is switched on at overflow and switched off at compare match,
    * so the active interval is from 0 to compare value.
    */
   stepSwitchOn();

//#ifdef TEST_MODE   
   if (work.time & 0x01) PORT_LED1 |= 1 << PIN_LED1;
   else                  PORT_LED1 &= (uint8_t) ~(1 << PIN_LED1);
   if (work.time & 0x02) PORT_LED2 |= 1 << PIN_LED2;
   else                  PORT_LED2 &= (uint8_t) ~(1 << PIN_LED2);
//#endif    

   if (!(work.cIPS % IO_CHECK)) {
      if (PORT_RIGHT_ENCODE & 1 << PIN_RIGHT_ENCODE) 
         i = 1;
      if (PORT_LEFT_ENCODE & 1 << PIN_LEFT_ENCODE) 
         i ^= 3;                                           // gray to binary
      i -= enc_last;

      if (i & 1) {
         enc_last += i;

         stepSetSpeed((i & 2) - 1);
      }

      if (PORT_PUSH_ENCODE & 1 << PIN_PUSH_ENCODE) work.btPE++;
      else                                         work.btPE = 0;

      if (PORT_T1 & 1 << PIN_T1) work.btT1++;
      else                       work.btT1 = 0;

      if (PORT_T2 & 1 << PIN_T2) work.btT2++;
      else                       work.btT2 = 0;

      if (PORT_T3 & 1 << PIN_T3) work.btT3++;
      else                       work.btT3 = 0;
   }
}


/*
 * compare match of timer 1 is pwm base
 */
ISR(TIMER1_COMPB_vect) {
   stepSwitchOff();
}


/*
 * overflow of timer 2 is whistle
 */
ISR(TIMER2_OVF_vect) {
   if (work.beep) {
      work.beep--;
      PORT_BUZZ ^= 1 << PIN_BUZZ;
   }
}

