/*
 * file:       LCD.c
 * purpose:    functions to handle LCD usage, with support for different interfaces
 *             It is possible to use more than one interface method (i.e. serial and direct).
 *             Pin declarations are normally placed in Config.h
 *             using direct-interface, it is possible to declare each pin separately, if no
 *             entire port can be used (define USE_SINGLE_PINS).
 *             The serial interface does not rely on any hardware support. Any pin can be used.
 * author:     Santiago
 * copyright:  2008 (c) Creative commons - by-nc-sa - some rights reserved
 *             see http://creativecommons.org/licenses/by-nc-sa/3.0/ for details
 */
#include <inttypes.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "Config.h"
#include "LCD.h"


#ifdef WANT_SAMPLE
static void lcdSample(void);
PROGMEM char defLine0[] = "LCD Module SRD-20 HP";
PROGMEM char defLine1[] = "20 Digits x 2 Lines.";
#endif
PROGMEM char clrLine[]  = "                    ";
#ifdef WANT_SERIAL_LCD
static uint8_t Bits[] = { 0x01, 0x08, 0x04, 0x02, 0x20, 0x40, 0x80, 0x10 };
#endif

/*
 * serial lcd-code by peter dannegger
 * via shift register 74HC164
 * connections:
 *    avr      -> lcd-driver 
 *    74HC164  -> lcd:
 *    ----------------
 *    3     ->    D4
 *    4     ->    D7
 *    5     ->    D6
 *    6     ->    D5
 *    10    ->    D1
 *    11    ->    D2
 *    12    ->    D3
 *    13    ->    D0
 */
static void lcdWriteByte(uint8_t d, uint8_t isData) {
#ifdef WANT_SERIAL_LCD
   uint8_t  i=0;

   /*
    * first the serial data ...
    */
   for (i=0; i < 8; i++) {
      PORT_LCD_SER_CLOCK &= (uint8_t) ~(1 << PIN_LCD_SER_CLOCK);

      if (d & Bits[i]) PORT_LCD_SER_DATA |= 1 << PIN_LCD_SER_DATA;
      else             PORT_LCD_SER_DATA &= (uint8_t) ~(1 << PIN_LCD_SER_DATA);

      PORT_LCD_SER_CLOCK |= 1 << PIN_LCD_SER_CLOCK;
   }
#endif   
   
   if (isData) {
#ifdef WANT_SERIAL_LCD
      PORT_LCD_SER_DATA |= 1 << PIN_LCD_SER_DATA;
#endif      
#ifdef WANT_DIRECT_LCD    
      PORT_LCD_RS       |= 1 << PIN_LCD_RS;
#endif      
   } else {
#ifdef WANT_SERIAL_LCD
      PORT_LCD_SER_DATA &= (uint8_t) ~(1 << PIN_LCD_SER_DATA);
#endif      
#ifdef WANT_DIRECT_LCD    
      PORT_LCD_RS       &= (uint8_t) ~(1 << PIN_LCD_RS);
#endif      
   }

#ifdef WANT_DIRECT_LCD    
   PORT_LCD_ENABLE   |= 1 << PIN_LCD_ENABLE;
#endif      
#ifdef WANT_SERIAL_LCD
   PORT_LCD_SER_SYNC |= 1 << PIN_LCD_SER_SYNC;
#endif

   /*
    * and now the parallel data ...
    */
#ifdef WANT_DIRECT_LCD    
#if(USE_SINGLE_PINS)
   if (d & 0x01) PORT_LCD_D0 |= 1 << PIN_LCD_D0;
   else          PORT_LCD_D0 &= (uint8_t) ~(1 << PIN_LCD_D0);
   if (d & 0x02) PORT_LCD_D1 |= 1 << PIN_LCD_D1;
   else          PORT_LCD_D1 &= (uint8_t) ~(1 << PIN_LCD_D1);
   if (d & 0x04) PORT_LCD_D2 |= 1 << PIN_LCD_D2;
   else          PORT_LCD_D2 &= (uint8_t) ~(1 << PIN_LCD_D2);
   if (d & 0x08) PORT_LCD_D3 |= 1 << PIN_LCD_D3;
   else          PORT_LCD_D3 &= (uint8_t) ~(1 << PIN_LCD_D3);
   if (d & 0x10) PORT_LCD_D4 |= 1 << PIN_LCD_D4;
   else          PORT_LCD_D4 &= (uint8_t) ~(1 << PIN_LCD_D4);
   if (d & 0x20) PORT_LCD_D5 |= 1 << PIN_LCD_D5;
   else          PORT_LCD_D5 &= (uint8_t) ~(1 << PIN_LCD_D5);
   if (d & 0x40) PORT_LCD_D6 |= 1 << PIN_LCD_D6;
   else          PORT_LCD_D6 &= (uint8_t) ~(1 << PIN_LCD_D6);
   if (d & 0x80) PORT_LCD_D7 |= 1 << PIN_LCD_D7;
   else          PORT_LCD_D7 &= (uint8_t) ~(1 << PIN_LCD_D7);
#else   
   PORT_LCD_DATA = d;
#endif
#endif
   _delay_us(45);
#ifdef WANT_SERIAL_LCD
   PORT_LCD_SER_SYNC &= (uint8_t) ~(1 << PIN_LCD_SER_SYNC);
   _delay_us(45);
#endif   
#ifdef WANT_DIRECT_LCD    
   PORT_LCD_ENABLE   &= (uint8_t) ~(1 << PIN_LCD_ENABLE);
#endif   

}


static void lcdCommand(uint8_t c) {
   lcdWriteByte(c, 0);
   if (c & 0x01 || c & 0x02)
      _delay_ms(16);
}


static void lcdData(uint8_t d) {
   lcdWriteByte(d, 1);
}


static void lcdEnable(uint8_t really) {
   if (really) lcdCommand(0x0C);
   else        lcdCommand(0x08);
}

void lcdInit(void) {
#ifdef WANT_SERIAL_LCD
   DDR_LCD_SER_CLOCK |= 1 << PIN_LCD_SER_CLOCK;
   DDR_LCD_SER_DATA  |= 1 << PIN_LCD_SER_DATA;
   DDR_LCD_SER_SYNC  |= 1 << PIN_LCD_SER_SYNC;
#endif
#ifdef WANT_DIRECT_LCD   
   DDR_LCD_ENABLE |= 1 << PIN_LCD_ENABLE;
   DDR_LCD_RS     |= 1 << PIN_LCD_RS;
#if(USE_SINGLE_PINS)
   DDR_LCD_D0     |= 1 << PIN_LCD_D0;
   DDR_LCD_D1     |= 1 << PIN_LCD_D1;
   DDR_LCD_D2     |= 1 << PIN_LCD_D2;
   DDR_LCD_D3     |= 1 << PIN_LCD_D3;
   DDR_LCD_D4     |= 1 << PIN_LCD_D4;
   DDR_LCD_D5     |= 1 << PIN_LCD_D5;
   DDR_LCD_D6     |= 1 << PIN_LCD_D6;
   DDR_LCD_D7     |= 1 << PIN_LCD_D7;
#else
   DDR_LCD_DATA    = 0xFF;
#endif   
#endif

   _delay_ms(16);
   lcdCommand(0x38);
   _delay_ms(5);
   lcdCommand(0x38);
   _delay_us(110);
   lcdCommand(0x38);
   lcdCommand(0x38);
   lcdCommand(0x08);
   lcdClear();
   lcdCommand(0x06);
   lcdEnable(1);
   lcdConfigure(1, 1, 1);                                  // 8bit bus, 2 lines, big font

#ifdef WANT_SAMPLE   
   lcdSample();

   _delay_ms(5000);
#endif   
}


void lcdClear(void) {
   lcdCommand(0x01);
}


void lcdConfigure(uint8_t use8Bits, uint8_t  twoLines, uint8_t bigFont) {
   uint8_t  cmd = 0x20;

   if (use8Bits) cmd |= 0x10;
   if (twoLines) cmd |= 0x08;
   if (bigFont)  cmd |= 0x04;

   lcdCommand(cmd);
   lcdCommand(0x0C);                                       // display on, cursor off
   lcdCommand(0x06);                                       // enable increment mode
}


void lcdWrite(uint8_t * s, uint8_t l) {
   do {
      lcdData(*s++);
   } while (--l > 0);
}


void lcdPrint(char * s) {
   while (*s) lcdData(*s++);
}


void lcdPrintLine(uint8_t line, PGM_P s) {
   uint8_t  b;
   uint8_t  i;

   lcdEnable(0);
   switch (line) {
   case 1:  lcdCommand(0x80); break;
   case 2:  lcdCommand(0xC0); break;
   case 3:
   case 4:  // not yet
      break;      
   }
   for (i=0; (b = pgm_read_byte_near(&clrLine[i])); i++)
      lcdData(b);

   switch (line) {
   case 1:  lcdCommand(0x80); break;
   case 2:  lcdCommand(0xC0); break;
   case 3:
   case 4:  // not yet
      break;      
   }

   for (i=0; (b = pgm_read_byte_near(&s[i])); i++)
      lcdData(b);
   lcdEnable(1);
}

#ifdef WANT_SAMPLE
static void lcdSample(void) {
   lcdPrintLine(1, defLine0);
   lcdPrintLine(2, defLine1);
}
#endif
